from django.contrib import admin
from django.db.models import Max
from enumfields.admin import EnumFieldListFilter

from arkindex.dataimport.models import DataFile, DataImport, Repository, Revision, Worker, WorkerVersion


class DataFileInline(admin.StackedInline):
    model = DataImport.files.through
    extra = 0


class WorkerRunInline(admin.StackedInline):
    model = DataImport.versions.through
    raw_id_fields = ('version', )
    extra = 0


class ElementInline(admin.StackedInline):
    model = DataImport.elements.through
    raw_id_fields = ('element', )
    readonly_fields = ('element', )
    extra = 0


class DataImportAdmin(admin.ModelAdmin):
    list_display = ('id', 'creator', 'corpus', 'state', 'mode')
    list_filter = [('mode', EnumFieldListFilter), ]
    fieldsets = (
        (None, {'fields': ('id', 'creator', 'corpus', 'state', 'mode', 'workflow')}),
        ('Elements filters', {
            'fields': ('element', 'element_type', 'folder_type', 'name_contains', 'best_class')
        }),
    )
    readonly_fields = ('id', 'workflow', 'state')
    raw_id_fields = ('element', )
    inlines = [DataFileInline, WorkerRunInline, ElementInline]

    def get_queryset(self, *args, **kwargs):
        return super().get_queryset(*args, **kwargs) \
            .prefetch_related('workflow__tasks') \
            .annotate(last_run=Max('workflow__tasks__run'))


class DataFileAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'corpus')
    list_filter = ('corpus', )
    fields = ('id', 'name', 'size', 'content_type', 'corpus')
    readonly_fields = ('id', 'size', )
    inlines = [DataFileInline, ]


class RevisionAdmin(admin.ModelAdmin):
    list_display = ('id', 'hash', 'message', 'author', 'repo')
    list_filter = ('repo', )
    fields = ('id', 'hash', 'message', 'author', 'repo')
    readonly_fields = ('id', )
    search_fields = ('hash', 'message', )


class WorkerInline(admin.StackedInline):
    model = Worker


class RepositoryAdmin(admin.ModelAdmin):
    list_display = ('id', 'url', 'type')
    fields = ('id', 'url', 'type', 'hook_token', 'provider_name')
    readonly_fields = ('id', )
    inlines = [WorkerInline, ]


class WorkerVersionInline(admin.StackedInline):
    model = WorkerVersion


class WorkerAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'slug', 'type', 'repository')
    field = ('id', 'name', 'slug', 'type', 'repository')
    readonly_fields = ('id', )
    inlines = [WorkerVersionInline, ]


class WorkerVersionAdmin(admin.ModelAdmin):
    list_display = ('id', 'worker', 'revision')
    list_filter = ('worker', )
    field = ('id', 'worker', 'revision', 'configuration')
    readonly_fields = ('id', )


admin.site.register(DataImport, DataImportAdmin)
admin.site.register(DataFile, DataFileAdmin)
admin.site.register(Revision, RevisionAdmin)
admin.site.register(Repository, RepositoryAdmin)
admin.site.register(Worker, WorkerAdmin)
admin.site.register(WorkerVersion, WorkerVersionAdmin)
