import logging
import uuid

from django.db import transaction
from django.db.models import Count, Q
from rest_framework import permissions, status
from rest_framework.exceptions import PermissionDenied, ValidationError
from rest_framework.generics import (
    CreateAPIView,
    GenericAPIView,
    ListAPIView,
    ListCreateAPIView,
    RetrieveUpdateDestroyAPIView,
)
from rest_framework.response import Response

from arkindex.documents.models import (
    Classification,
    ClassificationState,
    Corpus,
    Element,
    ElementPath,
    MLClass,
    Transcription,
)
from arkindex.documents.serializers.ml import (
    AnnotatedElementSerializer,
    ClassificationCreateSerializer,
    ClassificationMode,
    ClassificationSerializer,
    ClassificationsSelectionSerializer,
    ClassificationsSerializer,
    CountMLClassSerializer,
    ElementTranscriptionsBulkSerializer,
    TranscriptionBulkSerializer,
    TranscriptionCreateSerializer,
    TranscriptionSerializer,
)
from arkindex.images.models import Zone
from arkindex.project.filters import SafeSearchFilter
from arkindex.project.mixins import ACLMixin, CorpusACLMixin, DeprecatedMixin, SelectionMixin
from arkindex.project.permissions import IsVerified, IsVerifiedOrReadOnly
from arkindex.project.triggers import reindex_start
from arkindex.users.models import Role

logger = logging.getLogger(__name__)


class TranscriptionCreate(CreateAPIView):
    """
    Create a single transcription on an element
    The transcription zone is defined by the element it is attached
    """
    serializer_class = TranscriptionCreateSerializer
    permission_classes = (IsVerified, )
    openapi_overrides = {
        'description': 'Create a single manual transcription attached to an element',
        'operationId': 'CreateTranscription',
        'tags': ['transcriptions'],
    }

    def get_object(self):
        if not hasattr(self, 'element'):
            self.element = super().get_object()
        return self.element

    def get_queryset(self):
        if not self.request:
            return Element.objects.none()

        # We need to use the default database to avoid stale read
        # when creating a transcription on a newly created element
        return Element.objects.using('default').filter(
            corpus__in=Corpus.objects.writable(self.request.user)
        )

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request:
            context['element'] = self.get_object()
        return context

    def perform_create(self, serializer):
        ts = Transcription.objects.create(
            element=self.element,
            **serializer.validated_data
        )

        # Index in ES
        reindex_start(element=self.element, transcriptions=True, elements=True)
        return ts

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        obj = self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)
        return Response(
            # Use a single transcription serializer for the response
            TranscriptionSerializer(obj).data,
            status=status.HTTP_201_CREATED, headers=headers
        )


class TranscriptionEdit(ACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Retrieve or edit a transcription
    Edition and deletion are allowed for manual transcriptions
    Edition and deletion of non-manual transcriptions are allowed for the admin
    """
    serializer_class = TranscriptionSerializer
    permission_classes = (IsVerified, )
    openapi_overrides = {
        'tags': ['transcriptions'],
    }

    def get_queryset(self):
        if not self.request:
            return Transcription.objects.none()
        return Transcription.objects \
            .select_related('element__corpus') \
            .filter(element__corpus__in=Corpus.objects.readable(self.request.user))

    def check_object_permissions(self, request, transcription):
        super().check_object_permissions(request, transcription)
        if self.request.method in permissions.SAFE_METHODS:
            return

        role = Role.Contributor
        detail = 'A write access to transcription element corpus is required.'
        if transcription.worker_version_id:
            role = Role.Admin
            detail = 'Only admins can edit non-manual transcription.'
        if not self.has_access(transcription.element.corpus, role.value):
            raise PermissionDenied(detail=detail)


class ElementTranscriptionsBulk(CreateAPIView):
    """
    Insert multiple sub elements with their transcriptions on a common parent
    """
    serializer_class = ElementTranscriptionsBulkSerializer
    permission_classes = (IsVerified, )
    openapi_overrides = {
        'description': 'Create multiple sub elements with their transcriptions on a common parent in one transaction',
        'operationId': 'CreateElementTranscriptions',
        'tags': ['transcriptions'],
        'responses': {
            '201': {
                'description': 'Created',
                'content': {
                    'application/json': {
                        'examples': [{
                            'summary': 'return_elements = false',
                            'value': {
                                'element_type': 'string',
                                'worker_version': 'string',
                                'transcriptions': {
                                    'polygon': [[]],
                                    'text': 'string',
                                    'score': 0
                                },
                                'return_elements': False
                            }
                        }, {
                            'summary': 'return_elements = true',
                            'value': [{
                                'id': 'string',
                                'created': True
                            }]
                        }]
                    }
                }
            }
        }
    }

    def get_object(self):
        if not hasattr(self, 'element'):
            self.element = super().get_object()
            if not self.element.zone_id:
                raise ValidationError({'element': ['Cannot create transcriptions on an element without a zone.']})
        return self.element

    def get_queryset(self):
        if not self.request:
            return Element.objects.none()
        # Load elements and their zones from the default DB to avoid
        # a stale read issue when an element is created by a worker
        # and this endpoint is immediately used after to create transcriptions
        return Element.objects.filter(
            corpus__in=Corpus.objects.writable(self.request.user)
        ).using('default').select_related('zone')

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request:
            context['element'] = self.get_object()
        return context

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        annotations = self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)

        if serializer.validated_data['return_elements']:
            # Re-serialize the annotated sub-elements
            data = AnnotatedElementSerializer(annotations, many=True).data
        else:
            data = serializer.data

        return Response(data, status=status.HTTP_201_CREATED, headers=headers)

    @transaction.atomic
    def perform_create(self, serializer):
        elt_type = serializer.validated_data['element_type']
        worker_version = serializer.validated_data['worker_version']
        annotations = serializer.validated_data['transcriptions']

        # Retrieve existing zones from the DB
        image_id = self.element.zone.image_id
        existing_zones = {
            # Use WKB representation to compare existing zones
            # to avoid comparing references or slower coordinates
            polygon.wkb: zone_id
            for polygon, zone_id in Zone.objects.using('default').filter(image_id=image_id).values_list('polygon', 'id')
        }

        # Retrieve or create required zones
        new_zones = []
        for annotation in annotations:
            # Retrieve or create the zone on the image
            zone_id = existing_zones.get(annotation['polygon'].wkb)
            if not zone_id:
                new_zone = Zone(
                    id=uuid.uuid4(),
                    image_id=image_id,
                    polygon=annotation['polygon'],
                )
                zone_id = new_zone.id
                existing_zones[annotation['polygon'].wkb] = zone_id
                new_zones.append(new_zone)
            annotation['zone_id'] = zone_id

        Zone.objects.bulk_create(new_zones)

        # Retrieve or create elements to attach transcriptions to
        missing_elements = []
        missing_paths = []
        # List existing direct children (with their zones and transcriptions)
        children = {
            element.zone_id: element
            for element in Element
            .objects
            .get_descending(self.element.id)
            .using('default')
            .filter(
                zone__image_id=image_id,
                type=elt_type,
                paths__path__last=self.element.id
            )
            .only('id', 'zone_id')
        }
        # Load the paths immediately to avoid iterating over them for each element
        paths = list(self.element.paths.values_list('path', flat=True))
        if not paths:
            # Support top level elements, by adding an empty initial path to trigger loops below
            paths = [[]]
        next_path_ordering = self.element.get_next_order(elt_type)
        for annotation in annotations:
            # Look for a direct children with the right type and zone
            annotation['element'] = children.get(annotation['zone_id'])
            if not annotation['element']:
                annotation['element'] = Element(
                    id=uuid.uuid4(),
                    zone_id=annotation['zone_id'],
                    corpus_id=self.element.corpus_id,
                    type=elt_type,
                    name=next_path_ordering + 1
                )
                # Specify the annotated element has been created
                annotation['created'] = True
                children[annotation['zone_id']] = annotation['element']
                missing_elements.append(annotation['element'])
                for parent_path in paths:
                    new_path = parent_path + [self.element.id]
                    # Add the children to all of its parent paths
                    missing_paths.append(ElementPath(
                        element=annotation['element'],
                        path=new_path,
                        ordering=next_path_ordering
                    ))
                next_path_ordering += 1
        # Create missing elements and links in bulk
        Element.objects.bulk_create(missing_elements)
        ElementPath.objects.bulk_create(missing_paths)

        # Create transcriptions for the elements if they does not exist
        transcriptions = []
        for annotation in annotations:
            transcriptions.append(Transcription(
                element=annotation['element'],
                worker_version=worker_version,
                text=annotation['text'],
                score=annotation['score']
            ))
        # Create missing transcriptions in bulk
        Transcription.objects.bulk_create(transcriptions)

        # Index in ES
        reindex_start(element=self.element, transcriptions=True, elements=True)

        # Return ordered annotations
        return annotations


class TranscriptionBulk(CreateAPIView):
    '''
    Create multiple transcriptions at once on existing elements
    '''
    openapi_overrides = {
        'operationId': 'CreateTranscriptions',
        'tags': ['transcriptions'],
    }
    permission_classes = (IsVerified, )
    serializer_class = TranscriptionBulkSerializer


class CorpusMLClassList(CorpusACLMixin, ListCreateAPIView):
    """
    List available classes in a corpus with their distribution over elements
    """
    serializer_class = CountMLClassSerializer
    openapi_overrides = {
        'operationId': 'ListCorpusMLClasses',
        'description': (
            'List all available classes in a corpus with their'
            ' distribution over elements (best classes count)'
        ),
        'tags': ['classifications'],
    }
    filter_backends = [SafeSearchFilter]
    search_fields = ['name']
    permission_classes = (IsVerifiedOrReadOnly, )

    def get_queryset(self):
        if not self.request:
            return MLClass.objects.none()
        best_classification_filter = Q(
            # Keep non rejected best or validated classifications
            (
                (
                    Q(classifications__high_confidence=True)
                    & ~Q(classifications__state=ClassificationState.Rejected)
                )
                | Q(classifications__state=ClassificationState.Validated)
            )
        )
        return MLClass.objects.filter(corpus=self.get_corpus(self.kwargs['pk'])) \
            .annotate(nb_best=Count(
                'classifications',
                filter=best_classification_filter,
            )) \
            .order_by('name')

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request and self.request.method == 'POST':
            context['corpus_id'] = self.get_corpus(self.kwargs['pk'], role=Role.Contributor)
        return context


class MLClassList(DeprecatedMixin, ListAPIView):
    """
    List all available classes
    """
    openapi_overrides = {
        'operationId': 'ListMLClasses',
        'description': 'List available machine learning classes in all corpora.',
        'tags': ['classifications'],
    }
    deprecation_message = 'ListMLClasses is deprecated. Please use `ListCorpusMLClasses` instead.'


class ClassificationCreate(CreateAPIView):
    """
    Add a single classification on an element. Can be used by end users & machine learning tools.
    """
    serializer_class = ClassificationCreateSerializer
    queryset = Classification.objects.all()
    permission_classes = (IsVerified, )
    openapi_overrides = {
        'tags': ['classifications'],
    }

    def perform_create(self, serializer):
        if serializer.validated_data['worker_version'] is None:
            # A manual classification is immediately valid
            serializer.save(
                moderator=self.request.user,
                state=ClassificationState.Validated,
                high_confidence=True,
                confidence=1
            )
        else:
            # A classification produced by a tool is simply saved with its attributes
            serializer.save()


class ClassificationBulk(CreateAPIView):
    """
    Create multiple classifications at once on the same element with the same classifier.
    """
    serializer_class = ClassificationsSerializer
    permission_classes = (IsVerified, )
    openapi_overrides = {
        'tags': ['classifications'],
    }


class ManageClassificationsSelection(SelectionMixin, CorpusACLMixin, CreateAPIView):
    """
    Create multiple classifications at once for a list of selected elements for a specific corpus
    """
    serializer_class = ClassificationsSelectionSerializer
    permission_classes = (IsVerified, )
    openapi_overrides = {
        'description': 'Manage classifications for a list of selected elements for a specific corpus.',
        'tags': ['classifications']
    }

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        corpus_id = serializer.validated_data['corpus_id']
        corpus = self.get_corpus(corpus_id, role=Role.Contributor)

        mode = serializer.validated_data['mode']
        if mode == ClassificationMode.Create:
            return self.create(corpus, request, *args, **kwargs)
        elif mode == ClassificationMode.Validate:
            elements = self.get_selection(corpus_id)
            Classification.objects.filter(
                element__in=elements,
                state=ClassificationState.Pending,
                high_confidence=True
            ).update(state=ClassificationState.Validated)
            return Response(status=status.HTTP_201_CREATED)
        raise NotImplementedError

    def create(self, corpus, request, *args, **kwargs):
        ml_class = MLClass.objects.filter(id=request.data['ml_class']).first()

        elements = self.get_selection(corpus.id)
        existing_element_ids = set(Classification.objects.filter(
            element_id__in=elements,
            ml_class_id=ml_class.id,
            worker_version_id=None,
        ).values_list('element_id', flat=True))

        classifications = []
        for element in elements.exclude(id__in=existing_element_ids):
            classifications.append(Classification(
                element=element,
                ml_class=ml_class,
                moderator=self.request.user,
                state=ClassificationState.Validated,
                high_confidence=False,
                confidence=1
            ))
        Classification.objects.bulk_create(classifications)

        return Response(
            status=status.HTTP_201_CREATED
        )


class ClassificationModerationActionsMixin(GenericAPIView):
    permission_classes = (IsVerified, )
    serializer_class = ClassificationSerializer

    def get_queryset(self):
        return Classification.objects.filter(element__corpus__in=Corpus.objects.writable(self.request.user))

    def put(self, request, *args, **kwargs):
        instance = self.get_object()
        instance.moderator = self.request.user
        instance.state = self.moderation_action
        instance.save(update_fields=['moderator', 'state'])
        serializer = self.get_serializer(instance)
        return Response(serializer.data, status=status.HTTP_200_OK)


class ClassificationValidate(ClassificationModerationActionsMixin):
    """
    Validate a classification
    """
    moderation_action = ClassificationState.Validated
    openapi_overrides = {
        'operationId': 'ValidateClassification',
        'description': 'Validate an existing classification.',
        'tags': ['classifications'],
    }


class ClassificationReject(ClassificationModerationActionsMixin):
    """
    Reject a classification
    """
    moderation_action = ClassificationState.Rejected
    openapi_overrides = {
        'operationId': 'RejectClassification',
        'description': 'Reject an existing classification. Manual classifications will be deleted.',
        'tags': ['classifications'],
    }

    def put(self, request, *args, **kwargs):
        instance = self.get_object()

        if not instance.worker_version_id:
            # Delete manual classifications upon rejection
            instance.delete()
            return Response(None, status=status.HTTP_204_NO_CONTENT)

        instance.moderator = self.request.user
        instance.state = self.moderation_action
        instance.save(update_fields=['moderator', 'state'])
        serializer = self.get_serializer(instance)
        return Response(serializer.data, status=status.HTTP_200_OK)
