import uuid
from itertools import chain

from django.db.models import prefetch_related_objects

from arkindex.documents.models import Element, Entity, Transcription


def search_transcriptions_post(data):
    '''
    Search pages with a transcription
    List pages ordered by document
    '''
    transcription_ids = [r.meta.id for r in data]
    ts = Transcription.objects \
                      .filter(id__in=transcription_ids) \
                      .order_by('-score') \
                      .select_related('element__zone__image__server', 'worker_version')
    element_ids = list(ts.values_list('element_id', flat=True))
    all_parent_paths = Element.objects.get_ascendings_paths(*element_ids)
    for trans in ts:
        trans.parent_paths = all_parent_paths.get(trans.element_id, [])
    return ts


def search_transcriptions_filter_post(data, element_id):
    if not isinstance(element_id, uuid.UUID):
        element_id = uuid.UUID(element_id)
    return filter(
        lambda t: element_id in [p.id for p in chain(*t.parent_paths)],
        search_transcriptions_post(data)
    )


def search_entities_post(data):
    """
    Search entities containing query
    """
    entity_ids = [uuid.UUID(r.meta.id) for r in data]
    # Preserve ordering from ElasticSearch
    entities = {
        entity.id: entity
        for entity in Entity.objects.filter(id__in=entity_ids)
    }
    # Ignore entities from ES that do not exist in DB
    return list(filter(None, map(entities.get, entity_ids)))


def search_elements_post(data):
    """
    Search elements for a query
    """
    elt_ids = list(map(uuid.UUID, (result.meta.id for result in data)))
    if not elt_ids:
        return []
    tr_ids = [
        uuid.UUID(transcription.id)
        for result in data
        for transcription in result.meta.inner_hits.transcriptions
    ]
    tr_totals = {
        uuid.UUID(result.meta.id): result.meta.inner_hits.transcriptions.hits.total
        for result in data
    }

    transcriptions = {
        t.id: t
        for t in Transcription.objects.filter(id__in=tr_ids).select_related('worker_version')
    }

    elts_tr_ids = {
        uuid.UUID(result.meta.id): [
            uuid.UUID(transcription.id)
            for transcription in result.meta.inner_hits.transcriptions
        ] for result in data
    }

    elts_date_range = {
        uuid.UUID(result.meta.id): result.date_range
        if getattr(result, 'date_range', None) else {}
        for result in data
    }

    elts = list(Element.objects.filter(id__in=elt_ids).prefetch_related('corpus', 'type', 'zone__image__server'))
    # Preserve the ordering given by ElasticSearch
    ordered_elts = list(filter(None, map(lambda eid: next((e for e in elts if e.id == eid), None), elt_ids)))

    all_paths = Element.objects.get_ascendings_paths(*(e.id for e in ordered_elts))
    prefetch_related_objects(
        [element for paths in all_paths.values() for path in paths for element in path],
        'type',
    )

    for elt in ordered_elts:
        elt.transcriptions_results = list(filter(None, [transcriptions.get(tid) for tid in elts_tr_ids[elt.id]]))
        elt.total_transcriptions = tr_totals[elt.id]
        elt.parent_paths = all_paths.get(elt.id, [])
        date_range = elts_date_range.get(elt.id, {})
        elt.date_from = date_range['gte'] if 'gte' in date_range else None
        elt.date_to = date_range['lt'] if 'lt' in date_range else None

    return ordered_elts
