from unittest.mock import patch

from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, Element
from arkindex.project.tests import FixtureAPITestCase
from arkindex.project.tools import build_tree
from arkindex.users.models import Role


class TestDestroyElements(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.volume_type = cls.corpus.types.get(slug='volume')
        cls.vol = cls.corpus.elements.get(name='Volume 1')
        cls.surface = cls.corpus.elements.get(name='Surface A')
        cls.private_corpus = Corpus.objects.create(name='private', public=False)

    def test_element_destroy_verified_user(self):
        with self.assertNumQueries(0):
            response = self.client.delete(reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_element_destroy_acl(self):
        self.private_corpus.memberships.create(user=self.user, level=Role.Contributor.value)
        self.client.force_login(self.user)
        castle_story = self.private_corpus.elements.create(
            type=self.volume_type,
            name='Castle story'
        )
        with self.assertNumQueries(5):
            response = self.client.delete(reverse('api:element-retrieve', kwargs={'pk': str(castle_story.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(
            response.json(),
            {'detail': 'You do not have admin access to this element.'}
        )

    @patch('arkindex.project.triggers.tasks.element_trash.delay')
    def test_element_destroy(self, delay_mock):
        self.client.force_login(self.user)
        castle_story = self.corpus.elements.create(
            type=self.volume_type,
            name='Castle story'
        )
        self.assertEqual(self.corpus.elements.filter(id=castle_story.id).exists(), True)
        with self.assertNumQueries(7):
            response = self.client.delete(reverse('api:element-retrieve', kwargs={'pk': str(castle_story.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        args, kwargs = delay_mock.call_args
        self.assertEqual(len(args), 0)
        self.assertCountEqual(list(kwargs.pop('queryset')), list(self.corpus.elements.filter(id=castle_story.id)))
        self.assertDictEqual(kwargs, {
            'delete_children': True,
            'user_id': self.user.id,
            'description': 'Element deletion',
        })

    @patch('arkindex.project.triggers.tasks.element_trash.delay')
    def test_non_empty_element(self, delay_mock):
        """
        We can now delete a non-empty element
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.delete(reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        args, kwargs = delay_mock.call_args
        self.assertEqual(len(args), 0)
        self.assertCountEqual(list(kwargs.pop('queryset')), list(self.corpus.elements.filter(id=self.vol.id)))
        self.assertDictEqual(kwargs, {
            'delete_children': True,
            'user_id': self.user.id,
            'description': 'Element deletion',
        })

    def test_element_trash_children(self):
        # Check this volume has children
        children = Element.objects.get_descending(self.vol.id)
        ids = list(children.values_list('id', flat=True))
        self.assertEqual(len(ids), 24)
        self.assertEqual(len(Element.objects.exclude(id__in=ids)), 5)

        with self.assertExactQueries('element_trash_children.sql', params={'id': str(self.vol.id)}):
            Element.objects.filter(id=self.vol.id).trash()

        with self.assertRaises(Element.DoesNotExist):
            self.vol.refresh_from_db()

        self.assertFalse(Element.objects.filter(id__in=ids).exists())
        # Excluding the children, only the volume should have been deleted
        self.assertEqual(len(Element.objects.exclude(id__in=ids)), 4)

    def test_element_trash_without_children(self):
        children = Element.objects.get_descending(self.vol.id)
        ids = list(children.values_list('id', flat=True))
        self.assertEqual(len(ids), 24)

        with self.assertExactQueries('element_trash_no_children.sql', params={'id': str(self.vol.id)}):
            Element.objects.filter(id=self.vol.id).trash(delete_children=False)

        with self.assertRaises(Element.DoesNotExist):
            self.vol.refresh_from_db()

        # No children were deleted
        self.assertEqual(Element.objects.filter(id__in=ids).count(), 24)

    def test_element_trash_no_order_by(self):
        children = Element.objects.get_descending(self.vol.id)
        ids = list(children.values_list('id', flat=True))
        self.assertEqual(len(ids), 24)

        with self.assertExactQueries('element_trash_children.sql', params={'id': str(self.vol.id)}):
            Element.objects.filter(id=self.vol.id).order_by('name').trash()

        with self.assertRaises(Element.DoesNotExist):
            self.vol.refresh_from_db()

        self.assertFalse(Element.objects.filter(id__in=ids).exists())

    def test_element_trash_related_filter_bug(self):
        """
        This test should fail when a bug in ElementQuerySet.trash is fixed:
        Filtering by metadata, classification, selection, DataImportElement, or DataImport
        can cause elements to not actually be deleted as the related tables are deleted first.
        For example, deleting all elements with classifications will cause all classifications
        to be deleted, but not elements, because the QuerySet is re-evaluated
        and no elements will have classifications.
        """
        ml_class = self.corpus.ml_classes.create(name='Krabby')
        page1r, page1v, page2r = Element.objects.get_descending(self.vol.id).filter(type__slug='page')
        page1r.classifications.create(ml_class=ml_class, confidence=0.42)
        page1v.classifications.create(ml_class=ml_class, confidence=0.42)
        page2r.classifications.create(ml_class=ml_class, confidence=0.42)

        queryset = Element.objects.filter(classifications__isnull=False)

        # Ensure the pages are all in this queryset, and therefore should be deleted
        self.assertEqual(queryset.filter(id__in=[page1r.id, page1v.id, page2r.id]).count(), 3)

        with self.assertExactQueries('element_trash_ml_class.sql'):
            Element.objects.filter(classifications__isnull=False).trash()

        # Pages did not get deleted
        page1r.refresh_from_db()
        page1v.refresh_from_db()
        page2r.refresh_from_db()

        # Classifications have been deleted
        self.assertFalse(ml_class.classifications.exists())

    def test_element_trash_deep(self):
        """
        Tests the repetition of path deletions with deep hierarchies
        Builds a tree of elements A→B→C→D→E and deletes A with its children
        """
        elements = build_tree(
            {
                'B': 'A',
                'C': 'B',
                'D': 'C',
                'E': 'D',
            },
            type=self.corpus.types.create(slug='houndour', display_name='Houndour', folder=True),
            corpus=self.corpus,
        )

        with self.assertExactQueries('element_trash_deep.sql', params={'id': str(elements['A'].id)}):
            Element.objects.filter(id=elements['A'].id).trash()

        self.assertFalse(Element.objects.filter(id__in=[e.id for e in elements.values()]).exists())

    @patch('arkindex.project.triggers.tasks.element_trash.delay')
    def test_destroy_corpus_elements_requires_login(self, delay_mock):
        with self.assertNumQueries(0):
            response = self.client.delete(reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertFalse(delay_mock.called)

    @patch('arkindex.project.triggers.tasks.element_trash.delay')
    def test_destroy_corpus_elements_requires_writable(self, delay_mock):
        self.corpus.memberships.update(level=Role.Guest.value)
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.delete(reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertFalse(delay_mock.called)

    @patch('arkindex.project.triggers.tasks.element_trash.delay')
    def test_destroy_corpus_elements_empty(self, delay_mock):
        self.client.force_login(self.user)
        self.assertFalse(self.corpus.elements.filter(name='blablablabla').exists())

        with self.assertNumQueries(7):
            response = self.client.delete(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                QUERY_STRING='name=blablablabla',
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertFalse(delay_mock.called)

    @patch('arkindex.project.triggers.tasks.element_trash.delay')
    def test_destroy_corpus_elements(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.delete(reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        args, kwargs = delay_mock.call_args
        self.assertEqual(len(args), 0)
        self.assertCountEqual(list(kwargs.pop('queryset')), list(self.corpus.elements.all()))
        self.assertDictEqual(kwargs, {
            'delete_children': True,
            'user_id': self.user.id,
            'description': 'Element deletion',
        })

    @patch('arkindex.project.triggers.tasks.element_trash.delay')
    def test_destroy_corpus_elements_delete_children(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.delete(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                QUERY_STRING='delete_children=false'
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        args, kwargs = delay_mock.call_args
        self.assertEqual(len(args), 0)
        self.assertCountEqual(list(kwargs.pop('queryset')), list(self.corpus.elements.all()))
        self.assertDictEqual(kwargs, {
            'delete_children': False,
            'user_id': self.user.id,
            'description': 'Element deletion',
        })

    @patch('arkindex.project.triggers.tasks.element_trash.delay')
    def test_destroy_corpus_elements_rejected_filters(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.delete(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                QUERY_STRING='with_best_classes=True&with_has_children=True&best_class=True',
                HTTP_IF_MODIFIED_SINCE='Thu, 02 Apr 2099 13:37:42 GMT',
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            'If-Modified-Since': ['This parameter is not allowed in DELETE requests.'],
            'best_class': ['This parameter is not allowed in DELETE requests.'],
            'with_best_classes': ['This parameter is not allowed in DELETE requests.'],
            'with_has_children': ['This parameter is not allowed in DELETE requests.'],
        })

        self.assertFalse(delay_mock.called)

    @patch('arkindex.project.triggers.tasks.element_trash.delay')
    def test_destroy_element_children_requires_login(self, delay_mock):
        with self.assertNumQueries(0):
            response = self.client.delete(reverse('api:elements-children', kwargs={'pk': self.vol.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertFalse(delay_mock.called)

    @patch('arkindex.project.triggers.tasks.element_trash.delay')
    def test_destroy_element_children_requires_writable(self, delay_mock):
        self.corpus.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.delete(reverse('api:elements-children', kwargs={'pk': self.vol.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertFalse(delay_mock.called)

    @patch('arkindex.project.triggers.tasks.element_trash.delay')
    def test_destroy_element_children_empty(self, delay_mock):
        self.client.force_login(self.user)
        element = self.corpus.elements.create(type=self.volume_type, name='Lonely element')

        with self.assertNumQueries(8):
            response = self.client.delete(reverse('api:elements-children', kwargs={'pk': element.id}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertFalse(delay_mock.called)

    @patch('arkindex.project.triggers.tasks.element_trash.delay')
    def test_destroy_element_children(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.delete(reverse('api:elements-children', kwargs={'pk': self.vol.id}))
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        args, kwargs = delay_mock.call_args
        self.assertEqual(len(args), 0)
        self.assertCountEqual(
            list(kwargs.pop('queryset')),
            # Direct children of the volume
            list(Element.objects.get_descending(self.vol.id).filter(paths__path__last=self.vol.id)),
        )
        self.assertDictEqual(kwargs, {
            'delete_children': True,
            'user_id': self.user.id,
            'description': 'Element deletion',
        })

    @patch('arkindex.project.triggers.tasks.element_trash.delay')
    def test_destroy_element_children_delete_children(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.delete(
                reverse('api:elements-children', kwargs={'pk': self.vol.id}),
                QUERY_STRING='delete_children=false'
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        args, kwargs = delay_mock.call_args
        self.assertEqual(len(args), 0)
        self.assertCountEqual(
            list(kwargs.pop('queryset')),
            # Direct children of the volume
            list(Element.objects.get_descending(self.vol.id).filter(paths__path__last=self.vol.id)),
        )
        self.assertDictEqual(kwargs, {
            'delete_children': False,
            'user_id': self.user.id,
            'description': 'Element deletion',
        })

    @patch('arkindex.project.triggers.tasks.element_trash.delay')
    def test_destroy_element_children_rejected_filters(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.delete(
                reverse('api:elements-children', kwargs={'pk': self.vol.id}),
                QUERY_STRING='with_best_classes=True&with_has_children=True&best_class=True',
                HTTP_IF_MODIFIED_SINCE='Thu, 02 Apr 2099 13:37:42 GMT',
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            'If-Modified-Since': ['This parameter is not allowed in DELETE requests.'],
            'best_class': ['This parameter is not allowed in DELETE requests.'],
            'with_best_classes': ['This parameter is not allowed in DELETE requests.'],
            'with_has_children': ['This parameter is not allowed in DELETE requests.'],
        })

        self.assertFalse(delay_mock.called)

    @patch('arkindex.project.triggers.tasks.element_trash.delay')
    def test_destroy_element_parents_requires_login(self, delay_mock):
        with self.assertNumQueries(0):
            response = self.client.delete(reverse('api:elements-parents', kwargs={'pk': self.surface.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertFalse(delay_mock.called)

    @patch('arkindex.project.triggers.tasks.element_trash.delay')
    def test_destroy_element_parents_requires_writable(self, delay_mock):
        self.corpus.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.delete(reverse('api:elements-parents', kwargs={'pk': self.surface.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertFalse(delay_mock.called)

    @patch('arkindex.project.triggers.tasks.element_trash.delay')
    def test_destroy_element_parents_empty(self, delay_mock):
        self.client.force_login(self.user)
        element = self.corpus.elements.create(type=self.volume_type, name='Lonely element')

        with self.assertNumQueries(7):
            response = self.client.delete(reverse('api:elements-parents', kwargs={'pk': element.id}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertFalse(delay_mock.called)

    @patch('arkindex.project.triggers.tasks.element_trash.delay')
    def test_destroy_element_parents(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.delete(reverse('api:elements-parents', kwargs={'pk': self.surface.id}))
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        args, kwargs = delay_mock.call_args
        self.assertEqual(len(args), 0)
        self.assertCountEqual(
            list(kwargs.pop('queryset')),
            # Direct parents of the surface
            list(Element.objects.get_ascending(self.surface.id, recursive=False)),
        )
        self.assertDictEqual(kwargs, {
            'delete_children': True,
            'user_id': self.user.id,
            'description': 'Element deletion',
        })

    @patch('arkindex.project.triggers.tasks.element_trash.delay')
    def test_destroy_element_parents_delete_parents(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.delete(
                reverse('api:elements-parents', kwargs={'pk': self.surface.id}),
                QUERY_STRING='delete_children=false'
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        args, kwargs = delay_mock.call_args
        self.assertEqual(len(args), 0)
        self.assertCountEqual(
            list(kwargs.pop('queryset')),
            # Direct parents of the surface
            list(Element.objects.get_ascending(self.surface.id, recursive=False)),
        )
        self.assertDictEqual(kwargs, {
            'delete_children': False,
            'user_id': self.user.id,
            'description': 'Element deletion',
        })

    @patch('arkindex.project.triggers.tasks.element_trash.delay')
    def test_destroy_element_parents_rejected_filters(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.delete(
                reverse('api:elements-parents', kwargs={'pk': self.surface.id}),
                QUERY_STRING='with_best_classes=True&with_has_children=True&best_class=True',
                HTTP_IF_MODIFIED_SINCE='Thu, 02 Apr 2099 13:37:42 GMT',
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            'If-Modified-Since': ['This parameter is not allowed in DELETE requests.'],
            'best_class': ['This parameter is not allowed in DELETE requests.'],
            'with_best_classes': ['This parameter is not allowed in DELETE requests.'],
            'with_has_children': ['This parameter is not allowed in DELETE requests.'],
        })

        self.assertFalse(delay_mock.called)
