from django.urls import reverse
from rest_framework import status
from rest_framework.serializers import DateTimeField

from arkindex.dataimport.models import RepositoryType, WorkerVersion
from arkindex.documents.models import Corpus, Entity, MLClass
from arkindex.project.tests import FixtureAPITestCase
from arkindex_common.enums import EntityType, MetaType


class TestRetrieveElements(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.creds = cls.user.credentials.get()
        cls.repo = cls.creds.repos.get(type=RepositoryType.Worker)
        cls.rev = cls.repo.revisions.get()
        cls.vol = cls.corpus.elements.get(name='Volume 1')
        cls.private_corpus = Corpus.objects.create(name='private', public=False)
        cls.worker_version = WorkerVersion.objects.get(worker__slug='reco')

    def setUp(self):
        self.page = self.corpus.elements.get(name='Volume 1, page 1r')
        self.folio_metadata = self.page.metadatas.get()
        self.metadata = self.page.metadatas.create(
            name='meta',
            type=MetaType.Text,
            value='text_metadata'
        )

    def test_get_element(self):
        ml_class = MLClass.objects.create(name='text', corpus=self.corpus)
        classification = self.vol.classifications.create(worker_version=self.worker_version, ml_class=ml_class)

        response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'id': str(self.vol.id),
            'name': 'Volume 1',
            'type': 'volume',
            'corpus': {
                'id': str(self.corpus.id),
                'name': 'Unit Tests',
                'public': True,
            },
            'thumbnail_url': self.vol.thumbnail.s3_url,
            'thumbnail_put_url': None,
            'worker_version': None,
            'zone': None,
            'metadata': 0,
            'classifications': [
                {
                    'id': str(classification.id),
                    'confidence': None,
                    'high_confidence': False,
                    'state': 'pending',
                    'worker_version': str(self.worker_version.id),
                    'ml_class': {
                        'id': str(ml_class.id),
                        'name': 'text',
                    }
                }
            ]
        })

    def test_get_element_zone(self):
        self.assertIsNotNone(self.page.zone)
        response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.page.id)}))
        self.assertDictEqual(response.json()['zone'], {
            'id': str(self.page.zone_id),
            'image': {
                'id': str(self.page.zone.image.id),
                'path': 'img1',
                'url': 'http://server/img1',
                's3_url': None,
                'width': 1000,
                'height': 1000,
                'status': 'unchecked',
                'server': {
                    'display_name': 'Test Server',
                    'max_width': None,
                    'max_height': None,
                    'url': 'http://server'
                }
            },
            'polygon': [[0, 0], [0, 1000], [1000, 1000], [1000, 0], [0, 0]],
            'url': 'http://server/img1/0,0,1000,1000/full/0/default.jpg'
        })

    def test_get_element_thumbnail_folder(self):
        """
        Check getting an element only gives a thumbnail URL with folders
        """
        self.assertTrue(self.vol.type.folder)
        response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['thumbnail_url'], self.vol.thumbnail.s3_url)
        self.assertIsNone(data['thumbnail_put_url'])

        self.assertFalse(self.page.type.folder)
        response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertIsNone(data['thumbnail_url'])
        self.assertIsNone(data['thumbnail_put_url'])

    def test_get_element_thumbnail_put_internal(self):
        """
        Check getting an element only gives a thumbnail PUT URL to bots for folders
        """
        self.client.force_login(self.internal_user)

        self.assertTrue(self.vol.type.folder)
        response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['thumbnail_put_url'], self.vol.thumbnail.s3_put_url)

        self.assertFalse(self.page.type.folder)
        response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIsNone(response.json()['thumbnail_put_url'])

    def test_list_element_metadata(self):
        revision_metadata = self.page.metadatas.create(
            name='revmeta',
            type=MetaType.Text,
            value='text_metadata',
            revision=self.rev
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:element-metadata', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json(), [
            {
                'dates': [],
                'id': str(self.folio_metadata.id),
                'name': 'folio',
                'revision': None,
                'type': 'text',
                'value': '1r',
                'worker_version': None
            },
            {
                'id': str(self.metadata.id),
                'name': str(self.metadata.name),
                'type': self.metadata.type.value,
                'value': self.metadata.value,
                'dates': [],
                'revision': None,
                'worker_version': None
            },
            {
                'id': str(revision_metadata.id),
                'name': str(revision_metadata.name),
                'type': revision_metadata.type.value,
                'value': revision_metadata.value,
                'dates': [],
                'revision': {
                    "id": str(self.rev.id),
                    "date": DateTimeField().to_representation(value=self.rev.created),
                    "hash": self.rev.hash,
                    "message": self.rev.message,
                    "author": self.rev.author,
                    "commit_url": self.rev.commit_url,
                    "repo_id": str(self.repo.id)
                },
                'worker_version': None
            },
        ])

    def test_list_element_metadata_with_entity(self):
        # Link an entity to metadata, db requests number should remains same
        entity = Entity.objects.create(
            type=EntityType.Person,
            corpus=self.corpus,
            name='Marc',
            worker_version=self.worker_version,
        )
        self.metadata.entity = entity
        self.metadata.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:element-metadata', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.json()), 1)
        self.assertListEqual(response.json(), [
            {
                'dates': [],
                'id': str(self.folio_metadata.id),
                'name': 'folio',
                'revision': None,
                'type': 'text',
                'value': '1r',
                'worker_version': None,
            },
        ])

    def test_list_element_metadata_with_worker_version(self):
        self.folio_metadata.worker_version = self.worker_version
        self.folio_metadata.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:element-metadata', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.json()), 2)
        self.assertListEqual(response.json(), [
            {
                'dates': [],
                'id': str(self.folio_metadata.id),
                'name': 'folio',
                'revision': None,
                'type': 'text',
                'value': '1r',
                'worker_version': str(self.worker_version.id),
            },
            {
                'id': str(self.metadata.id),
                'name': str(self.metadata.name),
                'type': self.metadata.type.value,
                'value': self.metadata.value,
                'dates': [],
                'revision': None,
                'worker_version': None
            },
        ])

    def test_list_element_metadata_wrong_acl(self):
        self.page.corpus = self.private_corpus
        self.page.save()
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:element-metadata', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_list_element_metadata_wrong_element_id(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:element-metadata', kwargs={'pk': '12341234-1234-1234-1234-123412341234'}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
