from unittest.mock import patch

from arkindex.dataimport.models import WorkerVersion
from arkindex.documents.dates import DateType, InterpretedDate
from arkindex.project.elastic import ESElement
from arkindex.project.tests import FixtureAPITestCase
from arkindex_common.enums import MetaType


class TestESDocuments(FixtureAPITestCase):

    @patch('arkindex.documents.models.MetaData.get_dates')
    def test_build_search_index_dates(self, get_dates_mock):
        element = self.corpus.elements.create(
            type=self.corpus.types.get(slug='act'),
            name='sister act',
        )
        element.metadatas.create(type=MetaType.Date, name='date', value='something')

        # Test Upper bound
        get_dates_mock.return_value = [InterpretedDate(1420, 5, type=DateType.Upper), ]
        date_range = ESElement.from_model(element).to_dict().get('date_range')
        self.assertDictEqual(date_range, {'lt': '1420-05||+1M'})

        # Test Lower bound
        get_dates_mock.return_value = [InterpretedDate(1418, type=DateType.Lower), ]
        date_range = ESElement.from_model(element).to_dict().get('date_range')
        self.assertDictEqual(date_range, {'gte': '1418'})

        # Test with both Lower and Upper bound
        get_dates_mock.return_value = [
            InterpretedDate(1418, type=DateType.Lower),
            InterpretedDate(1428, type=DateType.Upper),
        ]
        date_range = ESElement.from_model(element).to_dict().get('date_range')
        self.assertDictEqual(date_range, {'gte': '1418', 'lt': '1428||+1y'})

        # Test an exact date
        get_dates_mock.return_value = [InterpretedDate(1666, 2, 3), ]
        date_range = ESElement.from_model(element).to_dict().get('date_range')
        self.assertDictEqual(date_range, {'gte': '1666-02-03', 'lt': '1666-02-03||+1d'})

    def test_index_reference_metadata(self):
        element = self.corpus.elements.get(name='Volume 1')
        element.metadatas.create(type=MetaType.Reference, name='ref.', value='123ABC')
        es_document = ESElement.from_model(element)
        self.assertCountEqual(es_document.references, ['123abc'])

    def test_children_no_zone(self):
        """
        Ensure elements without zones are ignored when indexing an element's children
        """
        page = self.corpus.elements.get(name='Volume 1, page 1r')
        self.assertTrue(page.transcriptions.exists())
        surface = self.corpus.elements.create(
            type=self.corpus.types.get(slug='surface'),
            name='/dev/null',
            zone=None,
        )
        surface.add_parent(page)
        surface.transcriptions.create(
            text='invisible transcription',
            worker_version=WorkerVersion.objects.get(worker__slug='reco'),
        )
        texts = [tr['text'] for tr in ESElement.from_model(page).to_dict()['transcriptions']]
        self.assertNotIn('invisible transcription', texts)
