import logging
import urllib.parse
from uuid import UUID

from django.conf import settings
from django.contrib.auth import login, logout
from django.contrib.auth.forms import PasswordResetForm
from django.contrib.auth.mixins import UserPassesTestMixin
from django.contrib.auth.tokens import default_token_generator
from django.contrib.contenttypes.models import ContentType
from django.core.mail import send_mail
from django.db.models import Count, Prefetch, Q
from django.shortcuts import get_object_or_404
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.http import urlsafe_base64_encode
from django.views.generic import RedirectView
from django_rq.queues import get_queue
from django_rq.settings import QUEUES
from rest_framework import status
from rest_framework.exceptions import AuthenticationFailed, NotFound, PermissionDenied, ValidationError
from rest_framework.generics import (
    CreateAPIView,
    ListAPIView,
    RetrieveAPIView,
    RetrieveDestroyAPIView,
    RetrieveUpdateDestroyAPIView,
    UpdateAPIView,
)
from rest_framework.permissions import SAFE_METHODS
from rest_framework.response import Response
from rest_framework.views import APIView
from rq.job import JobStatus

from arkindex.dataimport.models import Repository
from arkindex.documents.models import Corpus
from arkindex.project.mixins import ACLMixin
from arkindex.project.permissions import IsAuthenticatedOrReadOnly, IsVerified
from arkindex.users.models import Group, OAuthStatus, Right, Role, Scope, User, UserScope
from arkindex.users.providers import get_provider, oauth_providers
from arkindex.users.serializers import (
    EmailLoginSerializer,
    GenericMembershipSerializer,
    GroupSerializer,
    JobSerializer,
    MemberGroupSerializer,
    MembershipCreateSerializer,
    MembershipSerializer,
    NewUserSerializer,
    OAuthCredentialsSerializer,
    OAuthProviderClassSerializer,
    OAuthRetrySerializer,
    PasswordResetConfirmSerializer,
    PasswordResetSerializer,
    UserSerializer,
    UserTranskribusSerializer,
)
from arkindex.users.utils import get_max_level

logger = logging.getLogger(__name__)


class ProvidersList(ListAPIView):
    """
    List supported OAuth providers
    """
    permission_classes = (IsVerified, )
    serializer_class = OAuthProviderClassSerializer
    pagination_class = None
    openapi_overrides = {
        'operationId': 'ListOAuthProviders',
        'tags': ['oauth'],
    }

    def get_queryset(self):
        return list(filter(lambda p: p.enabled(), oauth_providers))


class CredentialsList(ListAPIView):
    """
    List all OAuth credentials for the authenticated user.
    """
    permission_classes = (IsVerified, )
    serializer_class = OAuthCredentialsSerializer
    openapi_overrides = {
        'tags': ['oauth'],
    }

    def get_queryset(self):
        qs = self.request.user.credentials.order_by('id')
        if 'status' in self.request.query_params:
            qs = qs.filter(status=self.request.query_params['status'])
        return qs


class CredentialsRetrieve(RetrieveDestroyAPIView):
    """
    Retrieve OAuth credentials.
    """
    permission_classes = (IsVerified, )
    serializer_class = OAuthCredentialsSerializer
    openapi_overrides = {
        'tags': ['oauth'],
    }

    def get_queryset(self):
        return self.request.user.credentials.order_by('id')

    def perform_destroy(self, instance):
        instance.provider_class(request=self.request, credentials=instance).disconnect()
        super().perform_destroy(instance)


class UserRetrieve(RetrieveUpdateDestroyAPIView):
    """
    Retrieve information about the authenticated user
    """
    permission_classes = (IsAuthenticatedOrReadOnly, )
    serializer_class = UserSerializer
    openapi_overrides = {
        'tags': ['users'],
    }

    def get_object(self):
        return self.request.user

    def retrieve(self, *args, **kwargs):
        if not self.request.user.is_authenticated:
            # Include features on GET to let the frontend still know the instance features when not logged in
            return Response(
                {
                    'detail': PermissionDenied.default_detail,
                    'features': settings.ARKINDEX_FEATURES,
                },
                status=status.HTTP_403_FORBIDDEN,
            )

        return super().retrieve(*args, **kwargs)

    def perform_destroy(self, instance):
        logout(self.request)


class UserCreate(CreateAPIView):
    serializer_class = NewUserSerializer
    openapi_overrides = {
        'operationId': 'Register',
        'description': 'Register as a new user.',
        'security': [],
        'tags': ['users'],
    }

    def post(self, *args, **kwargs):
        if not settings.ARKINDEX_FEATURES['signup']:
            raise ValidationError(['Registration has been disabled on this instance.'])
        return super().post(*args, **kwargs)

    def perform_create(self, serializer):
        user = serializer.save()
        login(self.request, user)

        activation_url = '{}?{}'.format(
            self.request.build_absolute_uri(reverse('api:user-token')),
            urllib.parse.urlencode({
                'email': user.email,
                'token': default_token_generator.make_token(user),
            }),
        )
        sent = send_mail(
            # Subject cannot have any newlines
            subject=''.join(render_to_string('registration/verification_email_subject.txt').splitlines()),
            message=render_to_string(
                'registration/verification_email.html',
                context={
                    'url': activation_url,
                },
                request=self.request,
            ),
            from_email=None,
            recipient_list=[user.email],
            fail_silently=True,
        )
        if sent == 0:
            logger.error('Failed to send registration email to {}'.format(user.email))


class UserEmailLogin(CreateAPIView):
    """
    Login using an email and a password
    """
    serializer_class = EmailLoginSerializer
    openapi_overrides = {
        'operationId': 'Login',
        'security': [],
        'tags': ['users'],
    }

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        try:
            user = User.objects.get_by_natural_key(serializer.validated_data['email'])
        except User.DoesNotExist:
            raise AuthenticationFailed

        if not user.has_usable_password() or not user.check_password(serializer.validated_data['password']):
            raise AuthenticationFailed

        login(self.request, user)

        headers = self.get_success_headers(serializer.data)
        return Response(
            UserSerializer(user).data,
            status=status.HTTP_201_CREATED,
            headers=headers,
        )


class UserEmailVerification(APIView):
    """
    Verify a user's email address
    """

    openapi_overrides = {
        'operationId': 'VerifyEmail',
        'tags': ['users'],
        'parameters': [
            {
                'name': 'email',
                'in': 'query',
                'description': 'E-mail to verify',
                'required': True,
                'schema': {
                    'type': 'string',
                    'format': 'email',
                },
            },
            {
                'name': 'token',
                'in': 'query',
                'description': 'Verification token',
                'required': True,
                'schema': {
                    'type': 'string',
                },
            },
        ],
    }

    def get(self, *args, **kwargs):
        if not all(arg in self.request.GET for arg in ('email', 'token')):
            raise ValidationError

        try:
            user = User.objects.get_by_natural_key(self.request.GET['email'])
        except User.DoesNotExist:
            raise AuthenticationFailed

        if not default_token_generator.check_token(user, self.request.GET['token']):
            raise AuthenticationFailed

        user.verified_email = True
        user.save()
        UserScope.objects.bulk_create([
            UserScope(user=user, scope=Scope.CreateIIIFImage),
            UserScope(user=user, scope=Scope.UploadS3Image),
        ])

        if user.has_usable_password():
            login(self.request, user)
            # A user with a verified email is granted a private trial corpus with admin rights
            if not Corpus.objects.admin(user=user).exists():
                corpus = Corpus.objects.create(
                    name='My Project',
                    description=f'Project for {user.email}'
                )
                corpus.memberships.create(
                    user=user,
                    level=Role.Admin.value
                )
                # Add default types
                corpus.create_default_types()

            # Redirtect to home page
            url = '/'
        else:
            url = reverse('password_reset_confirm', kwargs={
                'uidb64': urlsafe_base64_encode(str(user.id).encode()),
                'token': self.request.GET['token'],
            })

        return Response(
            status=status.HTTP_302_FOUND,
            headers={'Location': url},
        )


class PasswordReset(CreateAPIView):
    """
    Start a password reset flow
    """
    serializer_class = PasswordResetSerializer
    openapi_overrides = {
        'operationId': 'ResetPassword',
        'security': [],
        'tags': ['users'],
    }

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        form = PasswordResetForm(serializer.validated_data)

        if form.is_valid():
            form.save(
                request=self.request,
                use_https=self.request.is_secure(),
                token_generator=default_token_generator,
                email_template_name='registration/password_reset_email.html',
                subject_template_name='registration/password_reset_subject.txt',
            )
            headers = self.get_success_headers(serializer.data)
            return Response(
                status=status.HTTP_201_CREATED,
                headers=headers,
            )
        else:
            return Response(
                form.errors.get_json_data(),
                status=status.HTTP_400_BAD_REQUEST,
                headers=self.default_response_headers
            )


class PasswordResetConfirm(CreateAPIView):
    """
    Confirm a password reset using data from the confirmation e-mail
    """
    serializer_class = PasswordResetConfirmSerializer
    openapi_overrides = {
        'operationId': 'PasswordResetConfirm',
        'security': [],
        'tags': ['users'],
    }


class UserTranskribus(UpdateAPIView):
    """
    Update and validate Transkribus account
    Only the user's email is stored in our database
    """
    permission_classes = (IsVerified, )
    serializer_class = UserTranskribusSerializer
    openapi_overrides = {
        'tags': ['users'],
    }

    def get_object(self):
        return self.request.user


class OAuthSignIn(APIView):
    """
    Start the OAuth authentication code flow for a given provider
    """
    permission_classes = (IsVerified, )
    openapi_overrides = {
        'operationId': 'StartOAuthSignIn',
        'tags': ['oauth'],
    }

    def get(self, *args, **kwargs):
        if 'provider' not in kwargs:
            raise ValidationError('Missing provider')
        provider_class = get_provider(kwargs['provider'])
        if not provider_class:
            raise ValidationError('Unknown provider')
        url = self.request.GET.get('url', provider_class.url)
        if url:
            parsed = urllib.parse.urlparse(url)
            if (parsed.scheme != 'https'
                    or parsed.netloc == ''
                    or parsed.params != ''
                    or parsed.query != ''
                    or parsed.fragment != ''):
                raise ValidationError('Invalid GitLab instance URL')
        # Create OAuthCredentials without a token
        creds = self.request.user.credentials.create(
            provider_name=kwargs['provider'],
            provider_url=url,
        )
        return Response({
            'url': provider_class(request=self.request, credentials=creds).get_authorize_uri(),
        })


class OAuthRetry(RetrieveAPIView):
    """
    Retry the OAuth authentication code flow for pending credentials
    """
    permission_classes = (IsVerified, )
    serializer_class = OAuthRetrySerializer
    openapi_overrides = {
        'operationId': 'RetryOAuthCredentials',
        'tags': ['oauth'],
    }

    def get_queryset(self):
        return self.request.user.credentials

    def get_object(self):
        creds = super().get_object()
        if creds.status == OAuthStatus.Done:
            self.permission_denied()
        return creds

    def retrieve(self, request, *args, **kwargs):
        creds = self.get_object()
        return Response({
            'url': creds.provider_class(request=self.request, credentials=creds).get_authorize_uri(),
        })


class OAuthCallback(UserPassesTestMixin, RedirectView):
    """
    Callback for OAuth responses
    """
    pattern_name = 'credentials'
    permission_classes = (IsVerified, )
    raise_exception = True

    def test_func(self):
        """
        Perform Django REST framework permission checks, but without REST framework.
        """
        return all(perm().has_permission(self.request, self) for perm in self.permission_classes)

    def get(self, request, *args, **kwargs):
        assert 'provider' in kwargs
        provider_class = get_provider(kwargs['provider'])
        if not provider_class:
            raise ValueError('Unknown provider')
        provider = provider_class(self.request)
        try:
            provider.handle_callback()
            provider.credentials.status = OAuthStatus.Done
        except Exception as e:
            if not provider.credentials:
                raise
            provider.credentials.status = OAuthStatus.Error
            logger.warning(f'OAuth callback error: {e}', exc_info=e)
        provider.credentials.save()
        return super().get(request)


class JobList(ListAPIView):
    """
    List asynchronous jobs linked to the current user.
    """
    openapi_overrides = {
        'tags': ['jobs']
    }
    permission_classes = (IsVerified, )
    serializer_class = JobSerializer
    pagination_class = None

    def get_queryset(self):
        if not self.request:
            return []
        return self.request.user.get_rq_jobs()


class JobRetrieve(RetrieveDestroyAPIView):
    """
    Retrieve a single job by ID.
    """
    openapi_overrides = {
        'tags': ['jobs']
    }
    permission_classes = (IsVerified, )
    serializer_class = JobSerializer

    def get_object(self):
        for queue_name in QUEUES.keys():
            job = get_queue(queue_name).fetch_job(str(self.kwargs['pk']))
            if not job:
                continue
            if job.user_id != self.request.user.id:
                raise NotFound
            return job
        raise NotFound

    def perform_destroy(self, instance):
        if instance.get_status(refresh=False) == JobStatus.STARTED:
            raise ValidationError(['Cannot delete a running job.'])
        instance.delete()


class GroupsCreate(CreateAPIView):
    """
    Create a new group. The request user will be added as a member of this group.
    """
    serializer_class = GroupSerializer
    permission_classes = (IsVerified, )
    openapi_overrides = {
        'tags': ['users'],
    }


class GroupDetails(RetrieveUpdateDestroyAPIView):
    """
    Retrieve details about a specific group
    """
    serializer_class = GroupSerializer
    permission_classes = (IsVerified, )
    openapi_overrides = {
        'security': [],
        'tags': ['users'],
    }

    def get_membership(self, group):
        try:
            return group.memberships.get(user_id=self.request.user.id)
        except Right.DoesNotExist:
            raise PermissionDenied(detail='Only members of a group can retrieve their details')

    def get_object(self):
        if not hasattr(self, '_group'):
            self._group = super().get_object()
            if not hasattr(self, '_member'):
                self._member = self.get_membership(self._group)
            # Add request member level to the group
            self._group.level = self._member.level
        return self._group

    def check_object_permissions(self, request, obj):
        if not hasattr(self, '_member'):
            self._member = self.get_membership(obj)
        # Check the user has the right to delete or update a group
        super().check_object_permissions(request, obj)
        if request.method not in SAFE_METHODS and self._member.level < Role.Admin.value:
            raise PermissionDenied(detail='Only members with an admin privilege may update or delete this group.')

    def get_queryset(self):
        # Filter groups that are public or for which the user is a member
        return Group.objects \
            .annotate(members_count=Count('memberships')) \
            .filter(
                Q(public=True)
                | Q(memberships__user=self.request.user)
            )


class UserMemberships(ListAPIView):
    """
    List groups to which the request user belongs with its privileges on each group
    """
    serializer_class = MemberGroupSerializer
    permission_classes = (IsVerified, )
    openapi_overrides = {
        'security': [],
        'tags': ['users'],
    }

    def get_queryset(self):
        # Annotate rights with the group member count as Prefetch is not available with the Generic FK
        return self.request.user.rights \
            .prefetch_related('content_object') \
            .filter(content_type=ContentType.objects.get_for_model(Group)) \
            .annotate(members_count=Count('group_target__memberships')) \
            .order_by('group_target__name', 'group_target__id')


class MembershipDetails(RetrieveUpdateDestroyAPIView):
    """
    Retrieve details of a generic membership
    """
    serializer_class = MembershipSerializer
    permission_classes = (IsVerified, )
    openapi_overrides = {
        'tags': ['users'],
    }
    # Perform access checks with the permissions
    queryset = Right.objects.all()

    def check_object_permissions(self, request, membership):
        super().check_object_permissions(request, membership)

        # Check the access level of the request user on the content object
        access_level = get_max_level(self.request.user, membership.content_object)

        if not access_level or access_level < Role.Guest.value:
            # The request user does not have a read acces
            raise NotFound

        if (
            self.request.method not in SAFE_METHODS
            # Allow an user to remove its own membership
            and membership.user_id != self.request.user.id
            and access_level < Role.Admin.value
        ):
            raise PermissionDenied(
                detail='Only admins of the target membership group can perform this action.'
            )

    def perform_destroy(self, instance):
        if instance.level >= Role.Admin.value and instance.content_object.memberships.filter(level__gte=Role.Admin.value).count() < 2:
            raise ValidationError({'detail': ['Removing all memberships with an admin privilege is not possible.']})
        return super().perform_destroy(instance)


class GenericMembershipsList(ACLMixin, ListAPIView):
    """
    List memberships for a specific content with their privileges.
    Target of the right must be defined as a query parameter.
    """
    permission_classes = (IsVerified, )
    serializer_class = GenericMembershipSerializer
    content_object_params = {
        'group': Group,
        'corpus': Corpus,
        'repository': Repository
    }
    type_filters = ['user', 'group']
    openapi_overrides = {
        'tags': ['users'],
        'parameters': [
            {
                'name': param_name,
                'in': 'query',
                'description': f'List user members for a {param_name}',
                'required': False,
                'schema': {
                    'type': 'string',
                    'format': 'uuid',
                }
            }
            for param_name in content_object_params
        ] + [
            {
                'name': 'type',
                'in': 'query',
                'description': 'Filter memberships by owner type',
                'required': False,
                'schema': {
                    'enum': type_filters
                }
            }
        ]
    }

    def get_content_object(self):
        """
        Retrieve a generic content based on query parameters.
        This method does not perform any permission check.
        """
        content_keys = [key for key in self.request.query_params if key in self.content_object_params.keys()]
        # Assert only one of the content object parameter has been set
        if len(content_keys) != 1:
            raise ValidationError({
                '__all__': [
                    'Exactly one of those query parameters must be defined: '
                    f'{", ".join(self.content_object_params.keys())}.'
                ]
            })

        key, = content_keys
        content_id = self.request.query_params[key]
        try:
            content_uuid = UUID(content_id)
        except (AttributeError, ValueError):
            raise ValidationError({key: [f"'{content_id}' is not a valid UUID."]})

        # Use default database in case the content object has just been created
        return get_object_or_404(self.content_object_params[key].objects.using('default'), id=content_uuid)

    def get_queryset(self):
        content_object = self.get_content_object()
        # Raise a permission denied if the user has no access to the content object
        access_params = {}
        if isinstance(content_object, Group):
            # A user may not be able to list group members if he is not a member
            access_params['skip_public'] = True
        if not self.has_access(content_object, Role.Guest.value, **access_params):
            raise PermissionDenied(detail='You do not have the required access level to list members for this content.')

        # Avoid a stale read when adding/deletig a member
        qs = content_object.memberships \
            .using('default') \
            .prefetch_related(
                'user',
                Prefetch('group', queryset=Group.objects.annotate(members_count=Count('memberships')))
            ) \
            .order_by('user__display_name', 'group__name', 'id')

        # Handle a simple owner type filter
        type_filter = self.request.query_params.get('type')
        if type_filter and type_filter in self.type_filters:
            type_filter = {f'{type_filter}__isnull': False}
            qs = qs.filter(**type_filter)

        return qs


class GenericMembershipCreate(CreateAPIView):
    """
    Create a new generic membership.
    """
    permission_classes = (IsVerified, )
    serializer_class = MembershipCreateSerializer
    openapi_overrides = {
        'tags': ['users'],
    }
