import logging

from django.conf import settings
from django.contrib.auth.password_validation import validate_password
from django.contrib.auth.tokens import default_token_generator
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import transaction
from django.utils.http import urlsafe_base64_decode
from rest_framework import serializers

from arkindex.project.serializer_fields import EnumField
from arkindex.users.models import Group, OAuthCredentials, OAuthStatus, Right, Role, User
from arkindex.users.utils import RightContent, get_max_level
from transkribus import TranskribusAPI

logging.basicConfig(
    level=logging.INFO,
    format='[%(levelname)s] %(message)s',
)
logger = logging.getLogger(__name__)
transkribus_client = TranskribusAPI()


def validate_user_password(user, data):
    password = data.get('password')
    if password:
        try:
            validate_password(password, user)
        except ValidationError as e:
            raise serializers.ValidationError({'password': e.messages})
    return data


class OAuthCredentialsSerializer(serializers.ModelSerializer):

    provider_display_name = serializers.CharField(source='provider_class.display_name')
    status = EnumField(OAuthStatus)

    class Meta:
        model = OAuthCredentials
        fields = (
            'id',
            'status',
            'provider_name',
            'provider_display_name',
            'provider_url',
            'account_name',
        )


class OAuthProviderClassSerializer(serializers.Serializer):

    name = serializers.CharField(source='slug')
    display_name = serializers.CharField()
    default_url = serializers.URLField(source='url')


class OAuthRetrySerializer(serializers.Serializer):
    """
    A serializer used by the OAuthRetry endpoint to return an authorization URL.
    Required to get the OpenAPI schema generation to work.
    """

    url = serializers.URLField()


class SimpleUserSerializer(serializers.ModelSerializer):

    class Meta:
        model = User
        fields = ('id', 'email', 'display_name')


class UserSerializer(SimpleUserSerializer):

    features = serializers.SerializerMethodField(read_only=True)
    transkribus_import_email = serializers.SerializerMethodField(read_only=True)

    class Meta(SimpleUserSerializer.Meta):
        fields = SimpleUserSerializer.Meta.fields + (
            'password',
            'verified_email',
            'is_admin',
            'auth_token',
            'features',
            'transkribus_email',
            'transkribus_import_email',
        )
        extra_kwargs = {
            'id': {'read_only': True},
            'email': {'read_only': True},
            'password': {
                'write_only': True,
                'style': {
                    # Use a password input in DRF's HTML templates
                    'input_type': 'password',
                },
            },
            'verified_email': {'read_only': True},
            'is_admin': {'read_only': True},
            'auth_token': {'read_only': True},
            'transkribus_email': {'read_only': True},
        }

    def get_features(self, *args, **kwargs):
        return settings.ARKINDEX_FEATURES

    def get_transkribus_import_email(self, *args, **kwargs):
        return settings.TRANSKRIBUS_EMAIL

    def update(self, instance, validated_data):
        if 'password' in validated_data:
            instance.set_password(validated_data.pop('password'))
        return super().update(instance, validated_data)


class NewUserSerializer(serializers.ModelSerializer):
    """
    A serializer that allows non authenticated users to create an account with at
    least an email address
    """
    password = serializers.CharField(
        required=False,
        write_only=True,
        style={'input_type': 'password'}
    )

    class Meta:
        model = User
        fields = (
            'display_name',
            'email',
            'password'
        )

    def validate(self, data):
        data = super().validate(data)
        # Instanciate user to validate password
        user = User(**data)
        return validate_user_password(user, data)

    def create(self, validated_data):
        return User.objects.create_user(**validated_data)


class EmailLoginSerializer(serializers.Serializer):

    email = serializers.EmailField()
    password = serializers.CharField(write_only=True, style={'input_type': 'password'})


class PasswordResetSerializer(serializers.Serializer):
    """
    A serializer that only takes an e-mail as the input.
    This could use a ModelSerializer on User; however, just as the Django
    default password reset view does, there should not be any errors when
    asking for a password reset on a wrong e-mail, because that would give
    account existence information to attackers.
    """
    email = serializers.EmailField()


class PasswordResetConfirmSerializer(serializers.Serializer):
    """
    Second step of a password reset: when the user clicked the link in the email.
    The frontend will send the base64-encoded user ID, reset token and new password.
    """
    uidb64 = serializers.CharField(write_only=True)
    token = serializers.CharField(write_only=True)
    password = serializers.CharField(write_only=True)

    def validate(self, data):
        # Never fail, even with an invalid UUID or non-existent user
        try:
            pk = urlsafe_base64_decode(data['uidb64']).decode()
            user = User.objects.get(pk=pk)
        except (TypeError, ValueError, User.DoesNotExist):
            user = None

        valid_token = user is not None and default_token_generator.check_token(user, data['token'])

        validate_user_password(user, data)

        return {'user': user, 'valid_token': valid_token, 'password': data['password']}

    def save(self):
        user = self.validated_data['user']
        if not user or not self.validated_data['valid_token']:
            return
        user.set_password(self.validated_data['password'])
        user.save()


class UserTranskribusSerializer(serializers.Serializer):
    """
    A serializer that allows to verify Transkribus credentials
    and to save the Transkribus email
    """
    transkribus_email = serializers.EmailField()
    transkribus_password = serializers.CharField(write_only=True, style={'input_type': 'password'})

    def validate(self, data):
        transkribus_email = data.get('transkribus_email')
        transkribus_password = data.pop('transkribus_password')

        # Check that the credentials are correct
        try:
            transkribus_client.login(email=transkribus_email, password=transkribus_password)
        except Exception:
            logger.info("Invalid Transkribus email or password")
            raise serializers.ValidationError(
                {"__all__" : ["The email or password is incorrect"]}
            )

        return data

    def update(self, instance, validated_data):
        instance.transkribus_email = validated_data.get('transkribus_email', None)
        instance.save()
        return instance


class JobSerializer(serializers.Serializer):
    """
    Serializers a RQ job.
    """
    id = serializers.UUIDField(read_only=True)
    description = serializers.CharField(read_only=True)
    progress = serializers.FloatField(min_value=0, max_value=1, read_only=True, allow_null=True)
    status = serializers.SerializerMethodField()
    enqueued_at = serializers.DateTimeField(read_only=True, allow_null=True)
    started_at = serializers.DateTimeField(read_only=True, allow_null=True)
    ended_at = serializers.DateTimeField(read_only=True, allow_null=True)

    def get_status(self, instance):
        """
        Avoid causing more Redis queries to fetch a job's current status
        Note that a job status is part of a JobStatus enum,
        but the enum is just a plain object and not an Enum for Py2 compatibility.
        """
        return instance.get_status(refresh=False)


class SimpleGroupSerializer(serializers.ModelSerializer):
    """
    A light group serializer with its members count
    """
    members_count = serializers.IntegerField(default=None, read_only=True)

    class Meta:
        model = Group
        read_only_fields = ('id', 'members_count')
        fields = (
            'id',
            'name',
            'public',
            'members_count'
        )


class GroupSerializer(SimpleGroupSerializer):
    """
    A group serializer with the request member level
    """
    level = serializers.IntegerField(default=None, read_only=True)

    class Meta(SimpleGroupSerializer.Meta):
        fields = SimpleGroupSerializer.Meta.fields + ('level', )

    @transaction.atomic
    def create(self, validated_data):
        group = super().create(validated_data)
        # Associate the creator to the group
        group.add_member(self.context['request'].user, Role.Admin.value)
        # Manually set fields required by the serializer
        group.members_count = 1
        group.level = Role.Admin.value
        return group


class MembershipSerializer(serializers.ModelSerializer):
    """
    Simple serializer for a membership
    """

    class Meta:
        model = Right
        fields = (
            'id',
            'level',
        )

    def validate(self, data):
        data = super().validate(data)
        level_field = data.get('level')
        if not self.instance or not level_field:
            return data
        # Assert an update will not remove the last group admin
        if (self.instance.level >= Role.Admin.value
                and level_field < Role.Admin.value
                and self.instance.content_object.memberships.filter(level__gte=Role.Admin.value).count() < 2):
            raise ValidationError({'level': ['Removing all memberships with an admin privilege is not possible.']})
        return data


class MemberGroupSerializer(MembershipSerializer):
    """
    Serialize a group for a specific member with its privilege level
    """
    group = SimpleGroupSerializer(source='content_object')

    class Meta(MembershipSerializer.Meta):
        fields = MembershipSerializer.Meta.fields + ('group', )

    def __init__(self, instance=None, *args, **kwargs):
        if instance and isinstance(instance, list):
            # Annotate each target group with the memberships count
            for right in instance:
                right.content_object.members_count = right.members_count
        super().__init__(instance, *args, **kwargs)


class GenericMembershipSerializer(MembershipSerializer):
    """
    Serialize a generic membership
    """
    user = SimpleUserSerializer(allow_null=True)
    group = SimpleGroupSerializer(allow_null=True)

    class Meta(MembershipSerializer.Meta):
        fields = MembershipSerializer.Meta.fields + ('user', 'group')


class MembershipContentType(serializers.ChoiceField):
    """
    Specific field to handle a membership generic content
    """

    def __init__(self, *args, **kwargs):
        choices = [c.name for c in RightContent]
        super().__init__(choices, *args, **kwargs)

    def to_representation(self, obj):
        if not isinstance(obj, ContentType):
            return None
        model = obj.model_class()
        if model not in [c.value for c in RightContent]:
            return None
        return RightContent(model).name

    def to_internal_value(self, data):
        try:
            return ContentType.objects.get_for_model(RightContent[data].value)
        except Exception:
            raise ValidationError([f"'{data}' is not a valid content type."])


class MembershipCreateSerializer(MembershipSerializer):
    """
    Create a new generic membership
    """
    user_email = serializers.EmailField(source='user.email', required=False, allow_null=True)
    group_id = serializers.UUIDField(required=False, allow_null=True)
    content_type = MembershipContentType()
    content_id = serializers.UUIDField(help_text='The ID of the right target.')

    class Meta(MembershipSerializer.Meta):
        fields = (
            'id',
            'level',
            'user_email',
            'group_id',
            'content_type',
            'content_id'
        )

    def validate(self, data):
        data = super().validate(data)

        user_email = data.pop('user', {'email': None})['email']
        group_id = data.pop('group_id', None)

        # Assert one of user/group identifier is defined
        if not (user_email is None) ^ (group_id is None):
            raise ValidationError({'detail': ['Exactly one of those fields must be defined: user_email, group_id']})

        # Retrieve the right generic content object
        content_type = data.pop('content_type')
        content_model = content_type.model_class()
        content_id = data.pop('content_id')
        try:
            data['content_object'] = content_model.objects.get(id=content_id)
        except content_model.DoesNotExist:
            raise ValidationError({'content_id': [f'{content_type.name.capitalize()} with this ID could not be found.']})

        # Assert the user has any access to the content object
        max_user_level = get_max_level(self.context['request'].user, data['content_object'])
        if not max_user_level:
            raise ValidationError({'content_id': [f'{content_type.name.capitalize()} with this ID could not be found.']})

        # Special restriction for group memerships
        if content_model is Group and group_id is not None:
            raise ValidationError({'group_id': ['It is not possible to create a membership between two groups.']})

        # Retrieve owner of the right
        owner_data = {}
        if user_email:
            try:
                owner_data['user'] = User.objects.get(email=user_email)
            except User.DoesNotExist:
                raise ValidationError({'user_email': ['No user matching this email could be found.']})
        elif group_id:
            try:
                owner_data['group'] = Group.objects.get(id=group_id)
            except Group.DoesNotExist:
                raise ValidationError({'group_id': ['No group matching this id could be found.']})

        # Assert the request user a an admin access to the content object
        if max_user_level < Role.Admin.value:
            raise ValidationError({'content_id': ['Only members with an admin privilege are allowed to add other members.']})

        # Handle existing memberships
        if data['content_object'].memberships.filter(**owner_data).exists():
            raise ValidationError({'__all__': ['This membership already exists.']})

        data.update(owner_data)
        return data
