from django.db.models import IntegerField, Q, Value, functions
from django.db.models.query_utils import DeferredAttribute
from enumfields import Enum

from arkindex.dataimport.models import Repository
from arkindex.documents.models import Corpus
from arkindex.users.models import Group, Role

PUBLIC_LEVEL = Role.Guest.value


class RightContent(Enum):
    corpus = Corpus
    repository = Repository
    group = Group


def has_public_field(model):
    return isinstance(
        getattr(model, 'public', None),
        DeferredAttribute
    )


def get_public_instances(model):
    return model.objects \
        .filter(public=True) \
        .annotate(max_level=Value(PUBLIC_LEVEL, IntegerField()))


def check_level_param(level):
    assert type(level) is int, 'An integer level is required to compare access rights.'
    assert level >= 1, 'Level integer should be greater than or equal to 1.'
    assert level <= 100, 'level integer should be lower than or equal to 100'


def filter_rights(user, model, level):
    """
    Return a generic queryset of objects with access rights for this user.
    Level filtering parameter should be an integer between 1 and 100.
    """
    check_level_param(level)

    public = level <= PUBLIC_LEVEL and has_public_field(model)

    # Handle special authentications
    if user.is_anonymous:
        # Anonymous users have Guest access on public instances only
        if not public:
            return model.objects.none()
        return get_public_instances(model)
    elif user.is_admin or user.is_internal:
        # Superusers have an Admin access to all corpora
        return model.objects.all() \
            .annotate(max_level=Value(Role.Admin.value, IntegerField()))

    # Filter users rights and annotate the resulting level for those rights
    queryset = model.objects \
        .filter(
            # Filter instances with rights concerning this user
            Q(memberships__user=user)
            | Q(memberships__group__memberships__user=user)
        ) \
        .annotate(
            # Keep only the lowest level for each right via group
            max_level=functions.Least(
                'memberships__level',
                # In case of direct right, the group level will be skipped (Null value)
                'memberships__group__memberships__level'
            )
        ) \
        .filter(max_level__gte=level)

    # Use a join to add public instances as this is the more elegant solution
    if public:
        queryset = queryset.union(get_public_instances(model))

    return queryset


def get_max_level(user, instance):
    """
    Returns the maximum access level on a given instance
    """
    default_level = None
    if getattr(instance, 'public', False):
        default_level = PUBLIC_LEVEL

    # Handle special authentications
    if user.is_anonymous:
        return default_level
    elif user.is_admin or user.is_internal:
        # Superusers have an Admin access to all corpora
        return Role.Admin.value

    rights = instance.memberships \
        .filter(
            # Filter instances with rights concerning this user
            Q(user=user)
            | Q(group__memberships__user=user)
        ) \
        .annotate(
            # Keep only the lowest level for each right via group
            max_level=functions.Least(
                'level',
                # In case of direct right, the group level will be skipped (Null value)
                'group__memberships__level'
            )
        )

    # Filter out the right with the maximum level
    max_right = max(rights, key=lambda r: r.max_level, default=None)

    return max_right and max_right.max_level or default_level
